// Импорт модулей, функций
var fs = require('fs'),
    fsw = require('fswatch/lib/fswatch'),
    yfs = require('yanlibs/lib/yanfs'),
    yut = require('yanlibs/lib/yanutil'),
    fswapi = require('fswatch/lib/fswatchapi'),
    opts = yut.getopts(process.argv.slice(2, process.argv.length), ['--version', '-v', '-vv', '--help', '-h'], ['-w', '-e']).known,
    getMatcher = yfs.getMatcher,
    echoFile = yfs.echoFileSync;

// Пути и аргументы
var path = opts['-w'] || process.env['MJ_PWD'],
    pathl = path.length,
    evpath = opts['-e'],
    vverbose = opts['-vv'],
    home = process.env['MJ_HOME'],
    dirs = {};

if (opts['--version'] || opts['-v']) {
    echoFile(home + '/VERSION');
    process.exit();
}

if (opts['--help'] || opts['-h'] || !yfs.fileExistsSync(path + '/.mjoe.js')) {
    echoFile(home + '/USAGE');
    process.exit();
}

global.mjoe = {
    pwd: process.env['MJ_PWD'],
    exec: fsw.exec,
    E_CREATE: fswapi.CREATE,
    E_DELETE: fswapi.DELETE,
    E_MODIFY: fswapi.MODIFY,
    E_STATS_CHANGED: fswapi.STATS_CHANGED,
    E_ALL: fswapi.ALL
};

function unifyMatcher(m) {
    var t = m ? getMatcher(m) : function() { return false };
    return t instanceof RegExp ? function(s) { return t.test(s) } : t;
}

// Конфиг
var config = require(path + '/.mjoe').config,
    logpath = path + '/.mjoe.' + new Date().getTime() + '.log',
    exmatcher = unifyMatcher(config.exclude),
    preprocessor = config.preprocessor,
    workers = config.workers,
    interval = config.interval,
    maxdepth = config.depth || 20;

var ninotify = require('ninotify');

fsw.setLog(logmjoe);
fsw.registerParser(ninotify.parserid, ninotify);
fsw.registerWatcher(ninotify.watcherid, ninotify);
fsw.startQueue(interval);

function logmjoe(msg, outalso) {
    var fd = fs.openSync(logpath, 'a');
    msg = '[' + yut.dateToBasicString(new Date()) + '] ' + msg;
    outalso && console.log(msg);
    fs.writeSync(fd, msg + '\n', null);
    fs.closeSync(fd);
}

logmjoe.logpath = logpath;

function getdirs(dir, rdir, dirs, depth) {
    var content = fs.readdirSync(dir),
        p;

    if (depth > maxdepth) {
        logmjoe('Превышен лимит вложенности директорий: ' + dir, true);
        process.exit();
    }

    content.forEach(function(e) {
        p = dir + '/' + e;

        try {
            if (fs.lstatSync(p).isSymbolicLink()) p = fs.realpathSync(p);
        } catch (e) {
            logmjoe('Симлинк с ошибкой: ' + p, true);
            p = false;
        }

        if (p && fs.statSync(p).isDirectory() && !exmatcher(p)) {
            vverbose && console.log('Добавлена директория: ' + p);
            dirs[rdir + '/' + e] = p;
            getdirs(p, rdir + '/' + e, dirs, depth + 1);
        }
    });
}

var mjoecallback = function(e) {
    var ap = e.path, // абсолютный путь к событию
        rp = ap.substring(pathl, ap.length - 1); // относительный путь к событию

    if (e.mask && e.mask & 1) { // magic number: MJ_CREATE - временная мера
        try {
            if (fs.lstatSync(ap)) ap = fs.realpathSync(ap);
            if (fs.statSync(ap).isDirectory() && !exmatcher(rp)) {
                fsw.addPathAll(ap, evpath);
                vverbose && console.log('Добавлена директория: ' + rp);
            }
        } catch (e) {
            // Файла уже может и не быть на этот момент. Характерно для временных файлов vim'а
            vverbose && console.log(e);
        }
    }
};

logmjoe('Monkey Joe запускается...', true);

// собираем все директории
getdirs(path, '', dirs, 0);

// добавляем директории под присмотр, проверяя по исключению
for (var rdir in dirs) {
    fsw.addPathAll(dirs[rdir], evpath);
    vverbose && console.log('Добавлен watcher: ' + dirs[rdir]);
}

// подписываем пользовательские worker'ы на события
workers.forEach(function(w) {
    fsw.subscribe(evpath, w.matcher, preprocessor, w.callbacks);
});

// добавляем рутовую директорию конфига под присмотр
fsw.addPathAll(path, evpath);
fsw.subscribe(evpath, '*', null, mjoecallback);

logmjoe('Monkey Joe запустился', true);
